@@ -1,20 +1,24 @@
 """PDF stream."""
 
-import hashlib
 import io
 import struct
+from functools import lru_cache
+from hashlib import md5
 
 import pydyf
 from fontTools import subset
 from fontTools.ttLib import TTFont, TTLibError, ttFont
+from fontTools.varLib.mutator import instantiateVariableFont
 
 from ..logger import LOGGER
 from ..matrix import Matrix
-from ..text.ffi import ffi, harfbuzz, pango
+from ..text.ffi import ffi, harfbuzz, pango, units_to_double
 
 
 class Font:
-    def __init__(self, pango_font, hb_face):
+    def __init__(self, pango_font):
+        hb_font = pango.pango_font_get_hb_font(pango_font)
+        hb_face = harfbuzz.hb_font_get_face(hb_font)
         hb_blob = ffi.gc(
             harfbuzz.hb_face_reference_blob(hb_face),
             harfbuzz.hb_blob_destroy)
@@ -24,17 +28,21 @@ class Font:
         self.index = harfbuzz.hb_face_get_index(hb_face)
 
         pango_metrics = pango.pango_font_get_metrics(pango_font, ffi.NULL)
-        description = pango.pango_font_describe(pango_font)
-        font_size = pango.pango_font_description_get_size(description)
+        self.description = description = ffi.gc(
+            pango.pango_font_describe(pango_font),
+            pango.pango_font_description_free)
+        self.font_size = pango.pango_font_description_get_size(description)
         self.style = pango.pango_font_description_get_style(description)
         self.family = ffi.string(
             pango.pango_font_description_get_family(description))
-        digest = hashlib.sha1(self.file_content + bytes(self.index)).digest()
-        self.hash = ''.join(chr(65 + letter % 26) for letter in digest[:6])
-
-        # Name
         description_string = ffi.string(
             pango.pango_font_description_to_string(description))
+        # Never use the built-in hash function here: it’s not stable
+        self.hash = ''.join(
+            chr(65 + letter % 26) for letter
+            in md5(description_string).digest()[:6])
+
+        # Name
         fields = description_string.split(b' ')
         if fields and b'=' in fields[-1]:
             fields.pop()  # Remove variations
@@ -45,13 +53,13 @@ class Font:
         self.name = b'/' + self.hash.encode() + b'+' + b'-'.join(fields)
 
         # Ascent & descent
-        if font_size:
+        if self.font_size:
             self.ascent = int(
                 pango.pango_font_metrics_get_ascent(pango_metrics) /
-                font_size * 1000)
+                self.font_size * 1000)
             self.descent = -int(
                 pango.pango_font_metrics_get_descent(pango_metrics) /
-                font_size * 1000)
+                self.font_size * 1000)
         else:
             self.ascent = self.descent = 0
 
@@ -64,7 +72,10 @@ class Font:
             self.ttfont = None
             self.bitmap = False
         else:
-            self.bitmap = 'EBDT' in self.ttfont and 'EBLC' in self.ttfont
+            self.bitmap = (
+                'EBDT' in self.ttfont and
+                'EBLC' in self.ttfont and
+                not self.ttfont['glyf'].glyphs)
 
         # Various properties
         self.italic_angle = 0  # TODO: this should be different
@@ -81,7 +92,7 @@ class Font:
         self.flags = 2 ** (3 - 1)  # Symbolic, custom character set
         if self.style:
             self.flags += 2 ** (7 - 1)  # Italic
-        if b'Serif' in self.family.split():
+        if b'Serif' in fields:
             self.flags += 2 ** (2 - 1)  # Serif
         widths = self.widths.values()
         if len(widths) > 1 and len(set(widths)) == 1:
@@ -91,11 +102,13 @@ class Font:
         if self.ttfont is None:
             return
 
+        # Subset font
         if cmap:
             optimized_font = io.BytesIO()
             options = subset.Options(
                 retain_gids=True, passthrough_tables=True,
-                ignore_missing_glyphs=True, hinting=False)
+                ignore_missing_glyphs=True, hinting=False,
+                desubroutinize=True)
             options.drop_tables += ['GSUB', 'GPOS', 'SVG']
             subsetter = subset.Subsetter(options)
             subsetter.populate(gids=cmap)
@@ -107,6 +120,47 @@ class Font:
                 self.ttfont.save(optimized_font)
                 self.file_content = optimized_font.getvalue()
 
+        # Transform variable into static font
+        if 'fvar' in self.ttfont:
+            variations = pango.pango_font_description_get_variations(
+                self.description)
+            if variations == ffi.NULL:
+                variations = {}
+            else:
+                variations = {
+                    part.split('=')[0]: float(part.split('=')[1])
+                    for part in ffi.string(variations).decode().split(',')}
+            if 'wght' not in variations:
+                variations['wght'] = pango.pango_font_description_get_weight(
+                    self.description)
+            if 'opsz' not in variations:
+                variations['opsz'] = units_to_double(self.font_size)
+            if 'slnt' not in variations:
+                slnt = 0
+                if self.style == 1:
+                    for axe in self.ttfont['fvar'].axes:
+                        if axe.axisTag == 'slnt':
+                            if axe.maxValue == 0:
+                                slnt = axe.minValue
+                            else:
+                                slnt = axe.maxValue
+                            break
+                variations['slnt'] = slnt
+            if 'ital' not in variations:
+                variations['ital'] = int(self.style == 2)
+            partial_font = io.BytesIO()
+            try:
+                ttfont = instantiateVariableFont(self.ttfont, variations)
+                for key, (advance, bearing) in ttfont['hmtx'].metrics.items():
+                    if advance < 0:
+                        ttfont['hmtx'].metrics[key] = (0, bearing)
+                ttfont.save(partial_font)
+            except Exception:
+                LOGGER.warning('Unable to mutate variable font')
+            else:
+                self.ttfont = ttfont
+                self.file_content = partial_font.getvalue()
+
         if not (self.png or self.svg):
             return
 
@@ -140,16 +194,18 @@ class Font:
 class Stream(pydyf.Stream):
     """PDF stream object with extra features."""
     def __init__(self, fonts, page_rectangle, states, x_objects, patterns,
-                 shadings, images, *args, **kwargs):
+                 shadings, images, mark, *args, **kwargs):
         super().__init__(*args, **kwargs)
         self.compress = True
         self.page_rectangle = page_rectangle
+        self.marked = []
         self._fonts = fonts
         self._states = states
         self._x_objects = x_objects
         self._patterns = patterns
         self._shadings = shadings
         self._images = images
+        self._mark = mark
         self._current_color = self._current_color_stroke = None
         self._current_alpha = self._current_alpha_stroke = None
         self._current_font = self._current_font_size = None
@@ -170,7 +226,10 @@ class Stream(pydyf.Stream):
         self._ctm_stack.append(self.ctm)
 
     def pop_state(self):
-        super().pop_state()
+        if self.stream and self.stream[-1] == b'q':
+            self.stream.pop()
+        else:
+            super().pop_state()
         self._current_color = self._current_color_stroke = None
         self._current_alpha = self._current_alpha_stroke = None
         self._current_font = None
@@ -258,12 +317,18 @@ class Stream(pydyf.Stream):
             'BM': f'/{mode}',
         }))
 
+    @lru_cache()
     def add_font(self, pango_font):
-        hb_font = pango.pango_font_get_hb_font(pango_font)
-        hb_face = harfbuzz.hb_font_get_face(hb_font)
-        if hb_face not in self._fonts:
-            self._fonts[hb_face] = Font(pango_font, hb_face)
-        return self._fonts[hb_face]
+        if pango.pango_version() > 14600:
+            pango_face = pango.pango_font_get_face(pango_font)
+            description = pango.pango_font_face_describe(pango_face)
+        else:
+            description = pango.pango_font_describe(pango_font)
+        key = pango.pango_font_description_hash(description)
+        pango.pango_font_description_free(description)
+        if key not in self._fonts:
+            self._fonts[key] = Font(pango_font)
+        return self._fonts[key]
 
     def add_group(self, bounding_box):
         states = pydyf.Dictionary()
@@ -291,7 +356,7 @@ class Stream(pydyf.Stream):
         })
         group = Stream(
             self._fonts, self.page_rectangle, states, x_objects, patterns,
-            shadings, self._images, extra=extra)
+            shadings, self._images, self._mark, extra=extra)
         group.id = f'x{len(self._x_objects)}'
         self._x_objects[group.id] = group
         return group
@@ -426,7 +491,7 @@ class Stream(pydyf.Stream):
         })
         pattern = Stream(
             self._fonts, self.page_rectangle, states, x_objects, patterns,
-            shadings, self._images, extra=extra)
+            shadings, self._images, self._mark, extra=extra)
         pattern.id = f'p{len(self._patterns)}'
         self._patterns[pattern.id] = pattern
         return pattern
@@ -446,6 +511,22 @@ class Stream(pydyf.Stream):
         self._shadings[shading.id] = shading
         return shading
 
+    def begin_marked_content(self, box, mcid=False, tag=None):
+        if not self._mark:
+            return
+        property_list = None
+        if tag is None:
+            tag = self.get_marked_content_tag(box.element_tag)
+        if mcid:
+            property_list = pydyf.Dictionary({'MCID': len(self.marked)})
+            self.marked.append((tag, box))
+        super().begin_marked_content(tag, property_list)
+
+    def end_marked_content(self):
+        if not self._mark:
+            return
+        super().end_marked_content()
+
     @staticmethod
     def create_interpolation_function(domain, c0, c1, n):
         return pydyf.Dictionary({
@@ -465,3 +546,35 @@ class Stream(pydyf.Stream):
             'Bounds': pydyf.Array(bounds),
             'Functions': pydyf.Array(sub_functions),
         })
+
+    def get_marked_content_tag(self, element_tag):
+        if element_tag == 'div':
+            return 'Div'
+        elif element_tag == 'span':
+            return 'Span'
+        elif element_tag == 'article':
+            return 'Art'
+        elif element_tag == 'section':
+            return 'Sect'
+        elif element_tag == 'blockquote':
+            return 'BlockQuote'
+        elif element_tag == 'p':
+            return 'P'
+        elif element_tag in ('h1', 'h2', 'h3', 'h4', 'h5', 'h6'):
+            return element_tag.upper()
+        elif element_tag in ('dl', 'ul', 'ol'):
+            return 'L'
+        elif element_tag == 'li':
+            return 'LI'
+        elif element_tag == 'dt':
+            return 'LI'
+        elif element_tag == 'dd':
+            return 'LI'
+        elif element_tag == 'table':
+            return 'Table'
+        elif element_tag in ('tr', 'th', 'td'):
+            return element_tag.upper()
+        elif element_tag in ('thead', 'tbody', 'tfoot'):
+            return element_tag[:2].upper() + element_tag[2:]
+        else:
+            return 'NonStruct'
